from __future__ import unicode_literals
import json

from django.core.exceptions import ValidationError
from django.conf import settings
from django.db import models, DatabaseError, transaction
from django.utils.translation import ugettext_lazy as _
from django.utils import six
from django.core.cache import cache
from django.db import models

from decimal import Decimal
import datetime

from .utils import default
from .widgets import JSONWidget
from .forms import JSONFormField

from . import __version__

DB_TYPE_CACHE_KEY = (
    'django-jsonfield2:db-type:%s' % __version__ +
    '%(ENGINE)s:%(HOST)s:%(PORT)s:%(NAME)s'
)


class JSONField(six.with_metaclass(models.SubfieldBase, models.Field)):
    """
    A field that will ensure the data entered into it is valid JSON.
    """
    default_error_messages = {
        'invalid': _("'%s' is not a valid JSON string.")
    }
    description = "JSON object"

    __metaclass__ = models.SubfieldBase

    def __init__(self, *args, **kwargs):
        if not kwargs.get('null', False):
            kwargs['default'] = kwargs.get('default', dict)
        self.encoder_kwargs = {
            'indent': kwargs.pop('indent', getattr(settings, 'JSONFIELD_INDENT', None))
        }
        super(JSONField, self).__init__(*args, **kwargs)
        self.validate(self.get_default(), None)

    def formfield(self, **kwargs):
        defaults = {
            'form_class': JSONFormField,
            'widget': JSONWidget
        }
        defaults.update(**kwargs)
        return super(JSONField, self).formfield(**defaults)

    def validate(self, value, model_instance):
        if not self.null and value is None:
            raise ValidationError(self.error_messages['null'])
        try:
            self.get_prep_value(value)
        except ValueError:
            raise ValidationError(self.error_messages['invalid'] % value)

    def get_default(self):
        if self.has_default():
            default = self.default
            if callable(default):
                default = default()
            if isinstance(default, six.string_types):
                return json.loads(default)
            return json.loads(json.dumps(default))
        return super(JSONField, self).get_default()

    def get_internal_type(self):
        return 'TextField'

    def db_type(self, connection):
        if connection.vendor == 'postgresql':
            # Only do jsonb if in pg 9.4+
            if connection.pg_version >= 90400:
                return 'jsonb'
            return 'text'
        if connection.vendor == 'mysql':
            return 'longtext'
        return 'text'

    def to_python(self, value):
        if isinstance(value, six.string_types):
            if value == "":
                if self.null:
                    return None
                if self.blank:
                    return ""
            try:
                value = json.loads(value)
            except ValueError:
                msg = self.error_messages['invalid'] % value
                raise ValidationError(msg)
        # TODO: Look for date/time/datetime objects within the structure?
        return value

    def get_db_prep_value(self, value, connection=None, prepared=None):
        return self.get_prep_value(value)

    def get_prep_value(self, value):
        if value is None:
            if not self.null and self.blank:
                return ""
            return None
        return json.dumps(value, default=default, **self.encoder_kwargs)

    def get_prep_lookup(self, lookup_type, value):
        if lookup_type in ["exact", "iexact"]:
            return self.to_python(self.get_prep_value(value))
        if lookup_type == "in":
            return [self.to_python(self.get_prep_value(v)) for v in value]
        if lookup_type == "isnull":
            return value
        if lookup_type in ["contains", "icontains"]:
            if isinstance(value, (list, tuple)):
                raise TypeError("Lookup type %r not supported with argument of %s" % (
                    lookup_type, type(value).__name__
                ))
                # Need a way co combine the values with '%', but don't escape that.
                return self.get_prep_value(value)[1:-1].replace(', ', r'%')
            if isinstance(value, dict):
                return self.get_prep_value(value)[1:-1]
            return self.to_python(self.get_prep_value(value))
        raise TypeError('Lookup type %r not supported' % lookup_type)

    def value_to_string(self, obj):
        return self._get_val_from_obj(obj)


class TypedJSONField(JSONField):
    """

    """
    def __init__(self, *args, **kwargs):
        self.json_required_fields = kwargs.pop('required_fields', {})
        self.json_validators = kwargs.pop('validators', [])

        super(TypedJSONField, self).__init__(*args, **kwargs)

    def cast_required_fields(self, obj):
        if not obj:
            return
        for field_name, field_type in self.json_required_fields.items():
            obj[field_name] = field_type.to_python(obj[field_name])

    def to_python(self, value):
        value = super(TypedJSONField, self).to_python(value)

        if isinstance(value, list):
            for item in value:
                self.cast_required_fields(item)
        else:
            self.cast_required_fields(value)

        return value

    def validate(self, value, model_instance):
        super(TypedJSONField, self).validate(value, model_instance)

        for v in self.json_validators:
            if isinstance(value, list):
                for item in value:
                    v(item)
            else:
                v(value)

try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules([], ['^jsonfield2\.fields\.JSONField'])
    add_introspection_rules([], ['^jsonfield2\.fields\.TypedJSONField'])
except ImportError:
    pass



from django.db import models
from jsonfield2 import JSONField, JSONAwareManager

class JSONFieldTestModel(models.Model):
    json = JSONField("test", null=True, blank=True)

    class Meta:
        app_label = 'jsonfield2'


class JSONFieldWithDefaultTestModel(models.Model):
    json = JSONField(default={"sukasuka": "YAAAAAZ"})

    class Meta:
        app_label = 'jsonfield2'


class BlankJSONFieldTestModel(models.Model):
    null_json = JSONField(null=True)
    blank_json = JSONField(blank=True)

    class Meta:
        app_label = 'jsonfield2'


class CallableDefaultModel(models.Model):
    json = JSONField(default=lambda: {'x': 2})

    class Meta:
        app_label = 'jsonfield2'


class QueryJsonModel(models.Model):
    code = models.CharField( blank=False, null=False, max_length=20 )
    status = models.CharField( blank=True, null=True, max_length=20 )

    info = JSONField(default={})
    
    objects = JSONAwareManager(json_fields = ['info'])
    
    def __str__(self):
        return self.code

    class Meta:
        app_label = 'jsonfield2'



#:coding=utf-8:

from django.test import TestCase as DjangoTestCase
# from django.utils.encoding import force_text

from jsonfield2.tests.jsonfield_test_app.models import QueryJsonModel

class QJsonModelTest(DjangoTestCase):

    def setUp(self):

        data = [
            {"status": "A", "code": "lion", "info": "{}"},
            {"status": "0", "code": "cat", "info": "{}"},

            {"status": "A", "code": "Vito", "info": "{\"especie\": \"can\", \"sexo\": \"M\"}"},
            {"status": "0", "code": "Napa", "info": "{\"especie\": \"can\", \"sexo\": \"F\"}"}, 

            {"status": "A", "code": "Dario", "info": "{\"especie\": \"human\", \"sexo\": \"M\"}"},
            {"status": "A", "code": "Andrea", "info": "{\"especie\": \"human\", \"sexo\": \"F\"}"},
            {"status": "0", "code": "Xatli", "info": "{\"especie\": \"human\", \"sexo\": \"F\"}"}, 
        ]

        for reg in data:
            QueryJsonModel.objects.create(**reg)


    def test_q_nojson(self):
        rs = QueryJsonModel.objects.filter( status = 'A', info__icontains = 'human')    
        self.assertEqual(rs.count(), 2)

    def test_q_only_data(self):
        rs = QueryJsonModel.objects.filter( status = 'A')    
        self.assertEqual(rs.count(), 4)

    def test_q_only_json(self):
        rs = QueryJsonModel.objects.filter( info__especie = 'human')    
        self.assertEqual(rs.count(), 3)

    def test_q_data_json(self):
        rs = QueryJsonModel.objects.filter( status = 'A', info__especie = 'human')    
        self.assertEqual(rs.count(), 2)

    def test_q_data_json2(self):
        rs = QueryJsonModel.objects.filter( status = 'A', info__especie = 'human', info__sexo = 'M')
        self.assertEqual(rs.count(), 1)

    def test_s_only_data(self):
        rs = QueryJsonModel.objects.filter( status = 'A').order_by('code')
        self.assertEqual(rs.count(), 4)

    def test_s_only_json(self):
        rs = QueryJsonModel.objects.filter( info__especie = 'human').order_by('code')    
        self.assertEqual(rs.count(), 3)

    def test_s_data_json(self):
        rs = QueryJsonModel.objects.filter( status = 'A', info__especie = 'human').order_by('code')    
        self.assertEqual(rs.count(), 2)

    def test_s_data_json2(self):
        rs = QueryJsonModel.objects.filter( status = 'A', info__especie = 'human', info__sexo = 'M').order_by('code')
        self.assertEqual(rs.count(), 1)







import json

from django import template
from django.utils.safestring import mark_safe
from jsonfield2.utils import JSONEncoder

register = template.Library()


@register.filter
def jsonify(value):
    # If we have a queryset, then convert it into a list.
    if getattr(value, 'all', False):
        value = list(value)
    return mark_safe(json.dumps(value, cls=JSONEncoder))



import itertools, json 

from django.db import models
from django.utils import six

class JSONAwareQuerySet(models.query.QuerySet):
    def __init__(self, json_fields=[], *args, **kwargs):
        self.json_fields = json_fields
        super(JSONAwareQuerySet, self).__init__(*args, **kwargs)

    def _filter_or_exclude(self, negate, *args, **kwargs):
        extra_lookups = {}

        for lookup in kwargs:
            if lookup.split('__')[0] in self.json_fields:
                try:
                    lookupval = lookup.split('__')[1]
                except: 
                    continue 
                if lookupval in ["exact", "iexact", "in", "isnull", "contains", "icontains"]:
                    continue 
                extra_lookups[lookup] = kwargs[lookup]

        for key in extra_lookups:
            kwargs.pop(key)

        clone = super(JSONAwareQuerySet, self)._filter_or_exclude(negate, *args, **kwargs)

        if extra_lookups.keys():
            # self.json_lookups = True  
            len(clone)# Fill the cache

            # Lista de trabajo recorrida reversa para eliminar los no requeridos              
            resultlist = list( clone ) 

            # for item in itertools.product(self, extra_lookups.keys()):
            for i in range(len(resultlist)-1, -1, -1): 
                item = resultlist[ i ]
                for lookupkey in extra_lookups.keys():
                    evalresult = self._evaluate_json_lookup(item, lookupkey, extra_lookups[lookupkey])
                    if negate :  evalresult = not evalresult 
                    if not evalresult : 
                        resultlist.pop( i )
                        break

            clone._result_cache = resultlist

        return clone

    def _evaluate_json_lookup(self, item, lookup, value):
        oper = 'exact'

        evaluators = {
            'icontains': lambda item, value: item.lower() in value.lower(),
            'contains': lambda item, value: item in value,
            'in': lambda item, value: item in value,
            'iexact': lambda item, value: item.lower() == value.lower(),
            'exact': lambda item, value: item == value,
            'lt': lambda item, value: item < value,
            'lte': lambda item, value: item <= value,
            'gt': lambda item, value: item > value,
            'gte': lambda item, value: item >= value,
            'range': lambda item, value: item >= value[0] and item <= value[1],
        }

        def _getattr(obj, key):
            if isinstance(obj, dict):
                return obj[key]
            return getattr(obj, key)

        if lookup.split('__')[-1] in evaluators.keys():
            oper = lookup.split('__')[-1]
            lookup = '__'.join(lookup.split('__')[:-1])

        # DGT Verifica que el objeto json se un dictionario o lo convierte          
        field = getattr(item, lookup.split('__')[0])
        if isinstance(field, dict) : 
            jdict = field
        elif isinstance( field, ( six.string_types, six.text_type, bytes)) :
            try:
                jdict = json.loads(field)
            except :
                return False
        else: return False

        for key in lookup.split('__')[1:]:
            try:
                jdict = _getattr(jdict, key)
            except (AttributeError, KeyError):
                return False

        return evaluators[oper](jdict, value)


    def order_by(self, *args, **kwargs):
        # TODO: Dgt 1506 I d'nt find a way to know if the rs was filtered jsonfields
        # if self.json_lookups  : ...
        return self 

    def _clone(self, *args, **kwargs):
        clone = super(JSONAwareQuerySet, self)._clone(*args, **kwargs)
        clone.json_fields = self.json_fields
        # clone.json_lookups = getattr( self, 'json_lookups', False ) 

        return clone


class JSONAwareManager(models.Manager):
    
    def __init__(self, json_fields=[], *args, **kwargs):
        self.json_fields = json_fields
        super(JSONAwareManager, self).__init__(*args, **kwargs)

    def get_queryset(self):
        return JSONAwareQuerySet(self.json_fields, self.model)


# https://docs.djangoproject.com/en/1.8/topics/db/managers/#manager-names        


import datetime
import decimal
import types
import json
import uuid


from django.db.models.query import QuerySet
from django.utils import six, timezone
from django.utils.encoding import force_text
from django.utils.functional import Promise

from django.core.serializers.json import DjangoJSONEncoder

class JSONEncoder( DjangoJSONEncoder ):

    def default(self, obj):
        if isinstance(obj, Promise):
            return force_text(obj)

        elif isinstance(obj, datetime.timedelta):
            return six.text_type(total_seconds(obj))
 
        elif isinstance(obj, decimal.Decimal):
            # Serializers will coerce decimals to strings by default.
            return float(obj)
 
        elif isinstance(obj, uuid.UUID):
            return six.text_type(obj)
 
        elif isinstance(obj, QuerySet):
            return tuple(obj)
        elif hasattr(obj, 'tolist'):
            # Numpy arrays and array scalars.
            return obj.tolist()

        elif hasattr(obj, '__getitem__'):
            try:
                return dict(obj)
            except:
                pass
        elif hasattr(obj, '__iter__'):
            return tuple(item for item in obj)

        return super(JSONEncoder, self).default(obj)



def default(o):
    if hasattr(o, 'to_json'):
        return o.to_json()
    if isinstance(o, Decimal):
        return str(o)
    if isinstance(o, datetime.datetime):
        if o.tzinfo:
            return o.strftime('%Y-%m-%dT%H:%M:%S%z')
        return o.strftime("%Y-%m-%dT%H:%M:%S")
    if isinstance(o, datetime.date):
        return o.strftime("%Y-%m-%d")
    if isinstance(o, datetime.time):
        if o.tzinfo:
            return o.strftime('%H:%M:%S%z')
        return o.strftime("%H:%M:%S")
    if isinstance(o, set):
        return list(o)

    raise TypeError(repr(o) + " is not JSON serializable")



import os
from setuptools import setup

setup(
    name = "django-jsonfield2",
    version = '15.06.07',
    license='BSD',
   
    description='A flexible quering JSONField and associated form field to store validated JSON in your model.',
    long_description=open("README.rst").read(),

    install_requires=['Django >= 1.6'],
    tests_require=['Django >= 1.6'],
    
    url = "https://github.com/DarioGT/django-jsonfield2",
    author = "Dario Gomez-Tafur",
    author_email = 'certae.sm@gmail.com',
    
    packages = [
        "jsonfield2",
    ],
    classifiers = [
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    test_suite='runtest',
    include_package_data=True,
)
